module ARIParser where

import Term
import Rule
import TRS
import Text.ParserCombinators.Parsec
        
-- Scanners

comment :: Parser ()
comment = do
  _ <- char ';'
  skipMany (noneOf "\n")

whitespaces :: Parser ()
whitespaces = skipMany (do { _ <- space; return () } <|> comment)

identifier :: Parser String
identifier = do
  whitespaces
  x <- many1 (noneOf "(); \t\r\n")
  whitespaces
  return x

number :: Parser Int
number = do
  whitespaces
  s <- many1 digit
  whitespaces
  return (read s :: Int)

keyword :: String -> Parser ()
keyword s = do
  whitespaces
  _ <- string s
  whitespaces

param :: String -> a -> Parser a
param s x = do
  keyword s
  return x
            
-- Parsing functions.

parse_format :: Parser ()
parse_format = do
  keyword "("
  keyword "format"
  keyword "TRS"
  keyword ")"

parse_fun :: Parser (String, Int)
parse_fun = do
  keyword "("
  keyword "fun"
  f <- identifier
  n <- number
  keyword ")"
  return (f, n)

parse_term :: Signature -> Parser Term
parse_term sig = 
  try (parse_variable_or_constant sig) <|> 
  parse_function sig

parse_variable_or_constant :: Signature -> Parser Term
parse_variable_or_constant sig = do
  x <- identifier
  case lookup x sig of
    Nothing -> return (V x)
    Just 0  -> return (F x [])
    Just _  -> error (x ++ " is not a constant.")

parse_function :: Signature -> Parser Term
parse_function sig = do
  keyword "("
  f <- identifier
  ts <- many (parse_term sig)
  keyword ")"
  case lookup f sig of
    Nothing            -> error (f ++ " is not declared")
    Just n | m == n    -> return (F f ts)
           | m < n     -> error (f ++ " takes too few arguemnts")
           | otherwise -> error (f ++ " takes too many arguments")
      where m = length ts

parse_rule :: Signature -> Parser Rule
parse_rule sig = do
  keyword "("
  keyword "rule"
  l <- parse_term sig
  r <- parse_term sig
  keyword ")"
  return (l, r)

parse_ari :: Parser (Signature, TRS)
parse_ari = do
  parse_format
  sig <- many (try parse_fun)
  trs <- many (parse_rule sig)
  eof
  return (sig, trs)

-- A reader function.

read_file :: String -> IO (Signature, TRS)
read_file file = do
  result <- parseFromFile parse_ari file
  case result of
    Left e           -> error (show e)
    Right (sig, trs) -> return (sig, trs)
