module CPS where

import Data.List
import Term
import TRS
import Linear
import PCP
import Proof
import Proof2
import Rule
import SMT

-- { [t_i, t_{i-1},...,t1,s]) | s -> t1 -> .. -> ti for some i <= k }
at_most_k_steps :: Int -> TRS -> Term -> [([Term], [Rule])]
at_most_k_steps 0 _   s = [ ([s], []) ]
at_most_k_steps k trs s =
  nubBy eq (trss ++ [ (u : ts, nub (rule : rules)) | (ts@(t : _), rules) <- trss, (u, rule) <- reductsRules trs t ])
  where 
    eq ((v : _), _) ((w : _), _) = v == w
    eq _       _       = error "at_most_k_steps"
    trss = at_most_k_steps (k-1) trs s

joinSequence :: Int -> TRS -> Term -> Term -> Maybe [Term]
joinSequence k trs t u
  | t == u    = Just []
  | otherwise =
    case vss of
      vs : _ -> Just (init (tail vs))
      []     -> Nothing
    where
      trss = at_most_k_steps k trs t
      urss = at_most_k_steps k trs u
      vss = [ reverse ts ++ (v : us) | (v : ts, _) <- trss, (v' : us, _) <- urss, v == v' ]

rules_of_join_sequence :: Int -> TRS -> Term -> Term -> [TRS]
rules_of_join_sequence k trs t u
  | t == u    = [[]]
  | otherwise = [ nub (rules ++ rules') | (v : _, rules) <- trss, (v' : _, rules') <- urss, v == v']
  where trss = at_most_k_steps k trs t
        urss = at_most_k_steps k trs u

rulesOfToSequence :: Int -> TRS -> Term -> Term -> [TRS]
rulesOfToSequence k trs t u
  | t == u = [[]]
  | otherwise = [rules | (v : _, rules) <- at_most_k_steps k trs t, v == u]

joinable :: Int -> TRS -> Term -> Term -> Bool
joinable k trs t u = joinSequence k trs t u /= Nothing

pcpIsJoinable :: Int -> TRS -> Bool
pcpIsJoinable k trs = all p (pcpeak trs)
  where p (t,_,_,_,_,_,u) = joinable k trs t u

diagrams :: Int -> TRS -> [Peak] -> Maybe [CritPairInfo]
diagrams _ _    [] = Just []
diagrams k trs ((t, _, ps, _, s, _, u) : peaks) = do
    vs <- joinSequence k trs t u
    ds <- diagrams k trs peaks
    return ((t, ps, s, u, Nothing, vs) : ds)

pcps' :: Int -> TRS -> [Peak] -> ([CritPairInfo], TRS)
pcps' _ _ [] = ([], [])
pcps' k trsC ((t, rulesL, ps, _, s, ruleR, u) : peaks)
  | TRS.subsume trsC (ruleR : rulesL) =
      ((t, ps, s, u, Nothing, [s]) : joinSequences, trsP)
  | Just vs <- joinSequence k trsC t u =
      ((t, ps, s, u, Nothing, vs)  : joinSequences, trsP)
  | otherwise                          = 
      (joinSequences, (s, t) : (s, u) : trsP)
  where
    (joinSequences, trsP) = pcps' k trsC peaks

-- PCPS(R,C)
-- Instead of R, parallel critical peaks of R must be given. 
pcps :: Int -> [Peak] -> TRS -> ([CritPairInfo], TRS)
pcps k peaks trsC = (joinSequences, nub trsP)
  where (joinSequences, trsP) = pcps' k trsC peaks

rule_index' :: Int -> TRS -> Rule -> Maybe Int
rule_index' _ [] _ = Nothing
rule_index' i (rule' : rules) rule
  | Rule.variant rule rule' = Just i
  | otherwise               = rule_index' (i + 1) rules rule

rule_index :: TRS -> Rule -> Maybe Int
rule_index table rule = rule_index' 0 table rule

in_S :: TRS -> Rule -> Formula
in_S table rule
  | Just i <- rule_index table rule = FVar ("in_S_" ++ show i)
  | otherwise                       = error "in_S"

ruleVar' :: TRS -> Rule -> Int -> Maybe Int
ruleVar' [] _ _ = Nothing
ruleVar' (rule' : trs) rule n
  | Rule.variant rule rule' = Just n
  | otherwise               = ruleVar' trs rule (n + 1)

in_FunS :: String -> Formula
in_FunS f = FVar ("inFunS_" ++ f)

pIsEmpty :: Formula
pIsEmpty = FVar "pIsEmpty"

subset_of_S :: TRS -> TRS -> Formula
subset_of_S table trs = conj [ in_S table rule | rule <- trs ]

subset_of_FunS :: [String] -> Formula
subset_of_FunS fs = conj [ in_FunS f | f <- fs ]

encode_monotonicity :: Signature -> Formula
encode_monotonicity sig =
  conj 
    ([ Geq (var_fi f 0) (Val 0) | (f, _) <- sig ] ++
     [ Gt  (var_fi f i) (Val 0) | (f, n) <- sig, i <- [1..n] ])

encode_peaks :: Int -> TRS -> [Peak] -> Formula
encode_peaks k trs peaks =
  conj [ disj [subset_of_S table (rule : rules),
               disj [ subset_of_S table ss
                    | ss <- rules_of_join_sequence k trs t u ],
               conj [gtA s t, gtA s u, neg pIsEmpty]] 
       | (t, rules, _, _, s, rule, u) <- peaks ]
  where
    table = trs

encode_rule_removal :: Int -> TRS -> Formula
encode_rule_removal k trs =
  conj
    (neg (subset_of_S table trs) : 
     [ implies (in_S table rule) (subset_of_FunS (Rule.functions rule)) 
     | rule <- trs ] ++
     [ implies (subset_of_FunS (Term.functions l))
               (disj [ subset_of_S table ss
                     | ss <- rulesOfToSequence k trs l r])
     | (l, r) <- trs ])
  where table = trs

encode :: Int -> TRS -> [Peak] -> Formula
encode k trs peaks = 
  conj [encode_monotonicity (TRS.signature_of trs),
        implies (neg pIsEmpty) (conj [ geqA l r | (l, r) <- trs ]),
        encode_peaks k trs peaks,
        encode_rule_removal k trs]

decode_S :: TRS -> Model -> TRS
decode_S table model = 
  [ rule | rule <- table, eval_formula model (in_S table rule)]

decode_SN_proof :: Model -> Signature -> SNProof
decode_SN_proof model sig
  | eval_formula model pIsEmpty = RIsEmpty
  | otherwise                   = Linear (decode_algebra model sig)

reduce :: String -> Int -> TRS -> IO (Maybe (Subproof, TRS))
reduce smt k trs
  | TRS.leftLinear trs, 
    Just joinSequences <- diagrams k trs peaks = do
      m <- sat smt (CPS.encode k trs peaks)
      case m of
        Nothing    -> return Nothing
        Just model -> return (Just (subproof, subsystem))
          where
            subsystem = decode_S trs model
            (joinSequencesForC, trsP) = pcps k peaks subsystem
            subproof =
              CPS {
                __trsR = trs,
                __trsC = subsystem,
                __trsP = trsP,
                __joinSequencesForC = joinSequencesForC,
                __joinSequencesForR = joinSequences,
                __relativeTerminationProof = decode_SN_proof model (TRS.signature_of trs)
              }
  | otherwise = return Nothing
  where 
    peaks = pcpeak trs
