module CeTA where

import Data.List
import Term
import Rule
import TRS
import Proof
import Text.XML.Light
import Version
-- import TRSParser

type Attribute = (String, String)

type Criterion = TRS -> IO (Maybe CRResult)

type Certificate = (TRS, Signature, TRS, CRResult)


-- Labeling join sequences.

lookupBy :: (a -> a -> Bool) -> a -> [(a,b)] -> Maybe b
lookupBy _ _ [] = Nothing
lookupBy p x ((y, z) : a)
  | p x y     = Just z
  | otherwise = lookupBy p x a

label_of_rule :: Labeling a -> Rule -> a
label_of_rule phi rule
  | Just e <- lookupBy Rule.variant rule phi = e
  | otherwise = error "label_of_rule"

labels_of_rules:: Labeling a -> [Rule] -> [a]
labels_of_rules phi rules = [ label_of_rule phi rule | rule <- rules ]

-- XMLs

text :: String -> CData
text s = CData {cdVerbatim = CDataText, cdData = s, cdLine = Nothing}

int :: Int -> CData
int n = text (show n)

qname :: String -> String -> QName
qname prefix name = QName {qPrefix = Just prefix, qName = name, qURI = Nothing}

attr :: QName -> String -> Attr
attr key val = Attr {attrKey = key, attrVal = val}

-- term and rule

tag_term :: Term -> Element
tag_term (V x)    = unode "var" (text x)
tag_term (F f ts) = 
  unode "funapp" 
    (unode "name" (text f) :
     [tag_term t  | t <- ts ])

tag_rule :: Rule -> Element
tag_rule (l, r) = unode "rule" [tag_term l, tag_term r]

-- lookupTables

tag_index :: Int -> Element
tag_index i = unode "index" (int (i + 1))

tag_lookupTable :: TRS -> Element
tag_lookupTable trs =
  unode "lookupTables" 
    (unode "ruleTable" 
       [ unode "indexToRule" [tag_index i, tag_rule rule]
       | (i, rule) <- zip [0..] trs ])

-- input

tag_signature :: Signature -> Element
tag_signature sig =
  unode "signature"
    [ unode "symbol" [unode "name" (text f), unode "arity" (int n)]
    | (f, n) <- sig ]

tag_ruleIndex :: TRS -> Rule -> Element
tag_ruleIndex table rule
  | Just i <- elemIndex rule table = unode "ruleIndex" (int (i + 1))
  | otherwise                      = tag_rule rule

tag_rules :: TRS -> TRS -> Element
tag_rules table rules =
  unode "rules" [ tag_ruleIndex table rule | rule <- rules ]

tag_relativeRules :: TRS -> TRS -> Element
tag_relativeRules table rules =
  unode "relativeRules" [ tag_ruleIndex table rule | rule <- rules ]

tag_input :: TRS -> (Signature, TRS, TRS) -> Element
tag_input table (sig, rs, ss) =
  unode "input"
    (unode "trsWithSignature" 
       ([tag_signature sig, unode "trs" (tag_rules table rs)] ++ relatives))
  where
    relatives
      | ss == []  = []
      | otherwise = [tag_relativeRules table ss]


tag_trs :: TRS -> TRS -> Element
tag_trs table trs = unode "trs" (tag_rules table trs)

-- termination proofs based on linear interpretations

tag_monomial :: (Int, Int) -> Element
tag_monomial (0, c) = unode "integer" (int c)
tag_monomial (i, 1) = unode "variable" (int i)
tag_monomial (i, c) =
  unode "product"
    [unode "integer" (int c), 
     unode "variable" (int i)]

tag_sum :: [Element] -> Element
tag_sum []  = unode "integer" (int 0)
tag_sum [e] = e
tag_sum es  = unode "sum" es

tag_interpret :: Interpretation -> Element
tag_interpret (f, cs) =
  unode "interpret"
    [unode "name" (text f),
     unode "arity" (int (length cs - 1)),
     tag_sum [ tag_monomial (i, c) | (i, c) <- zip [0..] cs, c /= 0 ]]

tag_interpretation :: Algebra -> Element
tag_interpretation algebra =
  unode "interpretation"
    ([ unode "type" (unode "polynomial" [unode "domain" (unode "naturals" ()),
                                         unode "degree" (int 1)])] ++ 
     [ tag_interpret interpretation | interpretation <- algebra ])

tag_trsTerminationProof :: TRS -> TRS -> SNProof -> Element
tag_trsTerminationProof _ _ RIsEmpty =
  unode "trsTerminationProof" (unode "rIsEmpty" ())
tag_trsTerminationProof table trs (Linear algebra) =
  unode "trsTerminationProof"
    (unode "ruleRemoval" 
       [tag_interpretation algebra,
        unode "trs" (tag_rules table trs),
        tag_trsTerminationProof table [] RIsEmpty])

tag_terminationCertificationProblem :: (TRS, TRS, SNProof) -> Element
tag_terminationCertificationProblem (table, trs, snproof) = 
  add_attrs 
    [attr (qname "xmlns" "xsi") "http://www.w3.org/2001/XMLSchema-instance", 
     attr (qname "xsi" "noNamespaceSchemaLocation") "cpf3.xsd"]
    (unode "certificationProblem"
       [unode "cpfVersion" (text "3.0"),
        tag_lookupTable table,
        unode "input"
          (unode "trsInput"
             (unode "trs" (tag_rules table trs))),
        unode "property" (unode "termination" ()),
        unode "answer" (unode "yes" ()),
        unode "proof" (tag_trsTerminationProof table trs snproof),
        unode "metaInformation"
          (unode "toolInfos"
             [unode "toolInfo" (text "Hakusan"),
              unode "toolInfo" (text ("version: " ++ version))])])

show_terminationCertificate :: (TRS, TRS, SNProof) -> String
show_terminationCertificate certificate =
  -- "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n" ++
  "<?xml version=\"1.0\"?>\n" ++
  "<?xml-stylesheet type=\"text/xsl\" href=\"cpf3HTML.xsl\"?>\n" ++
  showElement (tag_terminationCertificationProblem certificate) ++ "\n"

-- rule labeling

tag_positionInTerm :: Position -> Element
tag_positionInTerm p = 
  unode "positionInTerm" 
    [ unode "position" (int (i + 1)) | i <- p ]

tag_labels :: Int -> Int -> Element
tag_labels maxLeft right =
  unode "labels" 
    [unode "maxLeft" (int maxLeft),
     unode "right" (int right)]

tag_intermediateTerms :: [Term] -> Element
tag_intermediateTerms ts =
  unode "intermediateTerms" [ tag_term t | t <- ts ]

tag_critPairInfo :: CritPairInfo -> Element
tag_critPairInfo (t, ps, s, u, labels, vs) = 
  unode "critPairInfo" 
    ([unode "left" (tag_term t), 
      unode "peak" (tag_term s), 
      unode "right" (tag_term u),
      unode "overlapPositions" [ tag_positionInTerm p | p <- ps ]] ++
      f labels ++
      [tag_intermediateTerms vs])
  where
    f Nothing = []
    f (Just (maxLeft, right)) = [tag_labels maxLeft right]

tag_joinSequences :: [CritPairInfo] -> Element
tag_joinSequences joinSequences =
  unode "joinSequences" 
    [ tag_critPairInfo critPairInfo | critPairInfo <- joinSequences ]


tag_ruleLabelingFunction :: TRS -> Labeling Int -> Element
tag_ruleLabelingFunction table phi =
  unode "ruleLabelingFunction"
    [ unode "ruleLabelingFunctionEntry" 
        [tag_ruleIndex table rule, unode "label" (int n)]
    | (rule, n) <- phi ]

index :: TRS -> Rule -> Int
index table rule
  | Just i <- elemIndex rule table = i
  | otherwise                = error "index"

indexes :: TRS -> TRS -> [Int]
indexes table trs = [ index table rule | rule <- trs ]

-- 

tag_answer :: CRResult -> Element
tag_answer (CRProof _)    = unode "answer" (unode "yes" ())
tag_answer (CRDisproof _) = unode "answer" (unode "no"  ())

tag_crProof :: TRS -> CRProof -> Element
tag_crProof table crProof = unode "crProof" (tag_crProof' table crProof)

tag_crProof' :: TRS -> CRProof -> Element
tag_crProof' _ EmptyTRS = unode "orthogonal" ()
tag_crProof' table r@(CompositionalPcpRuleLabeling {}) =
  unode "compositionalPcpRuleLabeling"
    [tag_ruleLabelingFunction table (_phi r),
     tag_ruleLabelingFunction table (_psi r),
     unode "joinsRS" (tag_joinSequences (_joinsRS r)),
     unode "joinsSR" (tag_joinSequences (_joinsSR r)),
     unode "trs" (tag_rules table (_trsC r)),
     tag_crProof table (_crProofForC r)]
tag_crProof' table r@(CompositionalPcps {}) =
  unode "compositionalPcps"
    [unode "trsC" (tag_rules table (_trsC r)),
     unode "trsP" (tag_rules table (_trsP r)),
     tag_joinSequences (_joinSequencesForC r),
     tag_joinSequences (_joinSequencesForR r),
     -- tag_relativeTerminationProof (_relativeTerminationProof r),
     tag_trsTerminationProof table (_trsP r) (_relativeTerminationProof r),
     tag_crProof table (_crProofForC r)]

tag_rewriteStep :: TRS -> [(Rule, Position, Term)] -> Element
tag_rewriteStep table steps =
  unode "rewriteStep"
    [ e | (rule, p, t) <- steps, 
          e <- [tag_positionInTerm p, tag_ruleIndex table rule, tag_term t] ]

tag_rewriteSequence :: TRS -> (Term,[(Rule,Position,Term)]) -> Element
tag_rewriteSequence table (s, steps) =
  unode "rewriteSequence"
    [unode "startTerm" (tag_term s),
     tag_rewriteStep table steps]

tag_nonJoinableFork :: TRS -> Fork -> Element
tag_nonJoinableFork table (t, rule1, p, s, rule2, u) =
  unode "nonJoinableFork" 
    [tag_rewriteSequence table (s, [(rule1, p, t)]),
     tag_rewriteSequence table (s, [(rule2, [], u)]),
     unode "capNotUnif" ()]

tag_crDisproof' :: TRS -> CRDisproof -> Element
tag_crDisproof' table (CapNotUnif fork) = 
  tag_nonJoinableFork table fork

tag_crDisproof :: TRS -> CRDisproof -> Element
tag_crDisproof table crDisproof = 
  unode "crDisproof" (tag_crDisproof' table crDisproof)

tag_proof' :: TRS -> CRResult -> Element
tag_proof' table (CRProof    crProof)    = tag_crProof    table crProof
tag_proof' table (CRDisproof crDisproof) = tag_crDisproof table crDisproof

tag_proof :: TRS -> CRResult -> Element
tag_proof table proof = unode "proof" (tag_proof' table proof)

tag_certificationProblem :: Certificate -> Element
tag_certificationProblem (table, sig, trs, proof) = 
  add_attrs 
    [attr (qname "xmlns" "xsi") "http://www.w3.org/2001/XMLSchema-instance", 
     attr (qname "xsi" "noNamespaceSchemaLocation") "cpf3.xsd"]
    (unode "certificationProblem"
       [unode "cpfVersion" (text "3.0"),
        tag_lookupTable table,
        tag_input table (sig, trs, []),
        unode "property" (unode "confluence" ()),
        tag_answer proof,
        tag_proof table proof,
        unode "metaInformation"
          (unode "toolInfos"
             [unode "toolInfo" (text "Hakusan"),
              unode "toolInfo" (text ("version: " ++ version))])])

show_certificate :: Certificate -> String
show_certificate certificate =
  -- "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n" ++
  "<?xml version=\"1.0\"?>\n" ++
  "<?xml-stylesheet type=\"text/xsl\" href=\"cpf3HTML.xsl\"?>\n" ++
  showElement (tag_certificationProblem certificate) ++ "\n"

-- for interface to TTT2

relativeTerminationProof :: String -> Maybe Element
relativeTerminationProof s = do
  root <- parseXMLDoc s
  proof <- findChild (unqual "proof") root
  findChild (unqual "trsTerminationProof") proof

tag_relativeTerminationProof :: String -> Element
tag_relativeTerminationProof proof =
  case relativeTerminationProof proof of
    Nothing -> error ("tag_relativeTerminationProof: Failed to parse\n" ++ proof)
    Just e  -> e

tag_wst_rule :: Rule -> Element
tag_wst_rule (l, r) =
  unode "rule" [unode "lhs" (tag_term l),
                unode "rhs" (tag_term r)]

tag_wst_relrules :: TRS -> Element
tag_wst_relrules trs = 
  unode "relrules" [ tag_wst_rule lr | lr <- trs ]

tag_wst_rules :: (TRS, TRS) -> Element
tag_wst_rules (rs, []) = 
  unode "rules" [ tag_wst_rule lr | lr <- rs ]
tag_wst_rules (rs, ss) =
  unode "rules" ([ tag_wst_rule lr | lr <- rs ] ++ [tag_wst_relrules ss])
 
tag_wst_trs :: (TRS, TRS) -> Element
tag_wst_trs (rs, ss) = 
  unode "trs" 
    [tag_wst_rules (rs, ss),
     tag_signature (TRS.signature_of (rs ++ ss))]

tag_problem :: (TRS, TRS) -> Element
tag_problem (rs, ss) = 
  add_attrs
    [attr (qname "xmlns" "xsi") "http://www.w3.org/2001/XMLSchema-instance",
     attr (unqual "type") "termination",
     attr (qname "xsi" "noNamespaceSchemaLocation") "/xtc.xsd"]
  (unode "problem" 
     [tag_wst_trs (rs, ss),
      unode "strategy" (text "FULL")])

show_problem :: (TRS, TRS) -> String
show_problem (rs, ss) = 
  "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n" ++
  showElement (tag_problem (rs, ss)) ++ "\n"
