module Main where

import Term
import TRS
import ARIParser
import qualified RL
import qualified CPS
import Linear
import qualified NonConfluenceCert
import Proof
import Proof2
import CeTA
import System.Environment

-- version
version :: String
version = "0.11"

-- main
help :: IO ()
help = do
  putStrLn "hakusan [option]* [file.ari]"
  putStrLn ("version " ++ version)
  putStrLn ""
  putStrLn "Options:"
  putStrLn "  -smt <s>  uses SMT solver <s>"
  putStrLn "  -R <k>    rule removal by rule labeling (Shintani and Hirokawa 2024)"
  putStrLn "  -C <k>    rule removal by critical pair systems (Shintani and Hirokawa 2024)"
  putStrLn "  -RC <k>   combination of R and C"
  putStrLn "  -v        version"
  putStrLn "  -h        help"
  putStrLn "Default setting is: -smt z3 -RC 5"

data Config = Config {
  _criterion :: Criterion,
  _smt :: String
}

showResult :: Signature -> TRS -> CRResult -> String
showResult sig trs proof@(CRProof _)    = 
  "YES\n" ++ CeTA.show_certificate (trs, sig, trs, proof)
showResult sig trs proof@(CRDisproof _) = 
  "NO\n"  ++ CeTA.show_certificate (trs, sig, trs, proof)

prove :: Criterion -> String -> IO ()
prove cr file = do 
  (sig, trs) <- read_file file
  m1 <- NonConfluenceCert.disprove trs
  case m1 of
    Just proof -> putStr (showResult sig trs proof)
    Nothing -> do
      m2 <- cr trs
      case m2 of
        Just proof -> putStr (showResult sig trs proof)
        Nothing    -> putStrLn "MAYBE"
        

type RuleRemoval = TRS -> IO (Maybe (Subproof, TRS))

apply :: [RuleRemoval] -> TRS -> IO (Maybe (Subproof, TRS))
apply [] _ = return Nothing
apply (criterion : criteria) trs = do
  m <- criterion trs
  case m of
    Nothing                    -> apply criteria trs
    Just (subproof, subsystem) -> return (Just (subproof, subsystem))

loop :: [RuleRemoval] -> TRS -> IO (Bool, [(Subproof, TRS)])
loop _        []  = return (True, [])
loop criteria trs = do
  m1 <- apply criteria trs
  case m1 of
    Nothing -> return (False, [])
    Just (subproof, subsystem) -> do
      (b, subproofs) <- loop criteria subsystem
      return (b, (subproof, subsystem) : subproofs)        

prove_confluence :: [RuleRemoval] -> TRS -> IO (Maybe CRResult)
prove_confluence criteria trs = do
  (b, subproofs) <- loop criteria trs
  if b then
    return (Just (CRProof (cr_proof subproofs)))
  else
    return Nothing

rl_cps :: String -> Int -> [RuleRemoval]
rl_cps smt k =
  (if k <= 2 then [] else [RL.reduce_all smt 2]) ++
  [RL.reduce_all smt k] ++
  (if k <= 2 then [] else [CPS.reduce smt 2]) ++
  [CPS.reduce    smt k,
   RL.reduce     smt k]

parseArgs :: Config -> [String] -> IO ()
parseArgs c ("-smt"   : s : args) =
  parseArgs (c { _smt = s }) args
parseArgs c ("-R" : sk : args) = parseArgs (c {_criterion = cr}) args
  where
    k = read sk
    cr trs
      | k <= 1    = prove_confluence [RL.reduce_all (_smt c) k] trs
      | otherwise = prove_confluence [RL.reduce_all (_smt c) 2,RL.reduce_all (_smt c) k] trs
parseArgs c ("-C" : sk : args) = parseArgs (c {_criterion = cr}) args
  where
    k = read sk
    cr trs
      | k <= 1    = prove_confluence [CPS.reduce (_smt c) k] trs
      | otherwise = prove_confluence [CPS.reduce (_smt c) 2,CPS.reduce (_smt c) k] trs
parseArgs c ("-RC" : sk : args) = parseArgs (c {_criterion = cr}) args
  where cr trs = prove_confluence (rl_cps (_smt c) (read sk)) trs
parseArgs c ["-l", file] = do
  (_, trs) <- read_file file
  m <- Linear.terminating (_smt c) trs []
  case m of
    Nothing -> putStrLn "MAYBE"
    Just snproof -> do
      putStrLn "YES"
      putStrLn (show_terminationCertificate (trs, trs, snproof))
parseArgs _ ["-h"] = help
parseArgs _ ["-v"] = putStrLn version
parseArgs c [file] = prove (_criterion c) file
parseArgs _ _ = help

defaultConfig :: Config
defaultConfig = 
  Config {
    _smt         = smt,
    _criterion   = prove_confluence (rl_cps smt 5)
  }
  where smt = "z3"

main :: IO ()
main = do
  args <- getArgs
  parseArgs defaultConfig args
